#include "config.h"
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "pixmap.h"
#include "tasklist.h"
#include "xhelp.h"
#include "atoms.h"
#include "globals.h"

static void FreeIcon(LPICON pIcon)
{
  if (pIcon->destroy)
  {
#ifdef DEBUG
    printf("Destroying icon 0x%08X\n",(unsigned int)pIcon->icon.icon);
#endif
    FreePixmap(&pIcon->icon);
  }
  pIcon->destroy = 0;
}

static void FreeTask(LPTASK task)
{
#ifdef DEBUG
  printf("Freeing task 0x%08X %s:%i\n",(unsigned int)task,task->name,task->nameLen);
#endif
  if (task->name)
  {
    free(task->name);
  }

  FreeIcon(&task->icon);

  free(task);
}

void ScaleIcon(LPTASK pTask)
{
  PIXMAP ret;  
  
  if (pTask->icon.icon.icon != None)
  {
    if (pTask->icon.icon.width != ICON_WIDTH || pTask->icon.icon.height != ICON_HEIGHT)
    {
      ret.width = ICON_WIDTH;
      ret.height = ICON_HEIGHT;
      ScalePixmap(pDisp,mainWindow,&pTask->icon.icon,&ret);
      if (pTask->icon.icon.icon != ret.icon)
      {
	FreeIcon(&pTask->icon);
	
	memcpy(&pTask->icon.icon,&ret,sizeof(PIXMAP));
	pTask->icon.destroy = 1;
      }
    }
  }
}

void LoadIcon(LPTASK pTask)
{
  /* try to load it from KWM */
  XWMHints *hin;
  unsigned long *data;

  /* free the current icon */
  FreeIcon(&pTask->icon);
  
  data = (unsigned long*)GetPropertyData(pTask->wnd,atom__KWM_WIN_ICON,atom__KWM_WIN_ICON,0);
  if (data)
  {
    pTask->icon.icon.icon = data[0];
    pTask->icon.icon.mask = data[1];
    UpdatePixmapAttributes(&pTask->icon.icon);
    /* ScaleIcon(pTask); */
    XFree(data);
  }
  else
  {
    /* now try loading a different way */
    hin = (XWMHints*)GetPropertyData(pTask->wnd,XA_WM_HINTS,XA_WM_HINTS,0);
    if (hin)
    {
      if (hin->flags & IconPixmapHint)
      {
	if (hin->flags & IconMaskHint)
	{
	  pTask->icon.icon.mask = hin->icon_mask;
	}
	pTask->icon.icon.icon = hin->icon_pixmap;
	UpdatePixmapAttributes(&pTask->icon.icon);
       	/* I should scale it now */
	ScaleIcon(pTask);
      }
      else			/* use defaults */
      {
	memcpy(&pTask->icon.icon,&defaultPixmap,sizeof(PIXMAP));
      }
      XFree(hin);
    }
    else			/* load the default one now */
    {
      memcpy(&pTask->icon.icon,&defaultPixmap,sizeof(PIXMAP));
    }
  }
}

LPTASK AddTask(LPTASK *head,Window newTask,unsigned int dwFlags)
{
  LPTASK tmp;
  LPTASK newObj;
  /* first determine if this is a top level window */
  if (GetDesktop(newTask) != gCurrentDesktop || IsHidden(newTask))
  {
    return NULL;
  }
  /* ok make a new task object now */
  newObj = (LPTASK)malloc(sizeof(TASK));
  if (!newObj)
  {
    return NULL;
  }
  
  memset(newObj,0,sizeof(TASK));
  newObj->wnd = newTask;
  newObj->dwFlags = dwFlags;
  /* newObj->icon = None;
  newObj->mask = None;
  newObj->next = NULL;
  newObj->name = NULL; */
  UpdateName(newObj);
  if (!newObj->name)
  {
    free(newObj);
    return NULL;
  }
  /* load the icons */
  LoadIcon(newObj);
  /* now listen for events */
  XSelectInput(pDisp,newTask,PropertyChangeMask | FocusChangeMask | StructureNotifyMask);
  /* now link up */
  if (*head)
  {
    tmp = *head;
    while (tmp->next)
    {
      tmp = tmp->next;
    }
    tmp->next = newObj;
  }
  else				/* set the head object */
  {
    *head = newObj;
  }
  return newObj;
}

void RemoveTask(LPTASK *head,Window deadTask)
{
  /* find it, then remove it */
  LPTASK tmp;
  LPTASK last=NULL;
  for (tmp = *head;tmp;tmp = tmp->next)
  {
    if (tmp->wnd == deadTask)
    {
      /* ok remove it here */
      if (last)
      {
	last->next = tmp->next;
      }
      else			/* this item must be the head */
      {
	*head = tmp->next;
      }
      FreeTask(tmp);
      return;
    }
    last = tmp;
  }  
}

void RemoveAllTasks(LPTASK *head)
{
  LPTASK next;
  LPTASK iter = *head;
  while (iter)
  {
    next = iter->next;
    /* remove head now */
    FreeTask(iter);
    /* now move to the next one */
    iter = next;
  }
  *head = NULL;
}

LPTASK FindTask(LPTASK head,LPFINDTASK pFindTask)
{
  for(;head;head = head->next)
  {
    if (pFindTask->dwFlags & FTF_WINDOW && head->wnd != pFindTask->wnd)
      continue;

    if (pFindTask->dwFlags & FTF_NAME && strcmp(head->name,pFindTask->name))
      continue;
    
    /* we fell through, we have a match */
    return head;
  }
  /* none found */
  return NULL;
}

void UpdateName(LPTASK pTask)
{
  if (pTask->name)
  {
    free(pTask->name);
    pTask->name = NULL;
    pTask->nameLen = 0;
  }
  
  pTask->name = GetWindowName(pTask->wnd,&pTask->nameLen);
}

unsigned int CountTasks(LPTASK head)
{
  unsigned int count=0;
  for (;head;head = head->next)
  {
    ++count;
  }
  return count;
}

LPTASK GetTask(LPTASK head,unsigned int index)
{
  unsigned int i=0;
  for (;head;head = head->next,++i)
  {
    if (index == i)
    {
      return head;
    }
  }
  return NULL;
}

void UnmarkAllTasks(LPTASK head)
{
  for (;head;head = head->next)
  {
    head->dwFlags &= ~TF_VALID;
  }
}

int RemoveUnmarkedTasks(LPTASK *head)
{
  LPTASK last = NULL;
  LPTASK header = *head;
  int numRemoved = 0;
  while (header)
  {
    if (!(header->dwFlags & TF_VALID))
    {
      /* remove this item now */
#ifdef DEBUG
      printf("Removing unmarked task\n");
#endif
      ++numRemoved;
      if (last)
      {
	last->next = header->next; /* link up */
	FreeTask(header);
	header = last->next;
      }
      else			/* update the head */
      {
	*head = header->next;
	FreeTask(header);
	header = *head;
      }
    }
    else
    {
      last = header;
      header = header->next;
    }
  }
  return numRemoved;
}

void MarkTask(LPTASK head)
{
  head->dwFlags |= TF_VALID;
}

void UnmarkTask(LPTASK head)
{
  head->dwFlags &= ~TF_VALID;
}
